#include <stdio.h> 
#include <windows.h> 

static const char PipeName[]="\\\\.\\pipe\\p-c-Pipe";

#define BUFSIZE 4096 

#define PIPE_TIMEOUT 4000

#define CONNECTING_STATE 0 
#define READING_STATE 1 
#define WRITING_STATE 2 
#define INSTANCES 4 

   CHAR chBuf[BUFSIZE]; 
   CHAR chBuf2[BUFSIZE]; 

   PROCESS_INFORMATION piProcInfo; 
   STARTUPINFO siStartInfo; 

   HANDLE hPipe;
   OVERLAPPED oOverlap; 
   BOOL fPendingIO;
   DWORD dwState; 
   DWORD dwMode;
 

VOID ErrorExit (LPTSTR lpszMessage) 
{ 
   fprintf(stderr, "%s\n", lpszMessage); 
   ExitProcess(0); 
} 
 

BOOL ConnectToNewClient(HANDLE hPipe, LPOVERLAPPED lpo) 
{ 
   BOOL fConnected, fPendingIO = FALSE; 
 
/* Start an overlapped connection for this pipe instance. */
   fConnected = ConnectNamedPipe(hPipe, lpo); 
 
/* Overlapped ConnectNamedPipe should return zero. */
   if (fConnected) 
    ErrorExit("Error: Connect Named Pipe (overlapped is not zero)"); 
 
   switch (GetLastError()) 
   { 
   /* The overlapped connection in progress. */
      case ERROR_IO_PENDING: 
         fPendingIO = TRUE; 
         break; 
 
   /* Client is already connected, so signal an event. */
 
      case ERROR_PIPE_CONNECTED: 
         if (SetEvent(lpo->hEvent)) 
            break; 
 
   /* If an error occurs during the connect operation... */
      default: 
         ErrorExit("Error: Connect Named Pipe"); 
   } 
 
   return fPendingIO; 
} 

VOID DisconnectAndReconnect() 
{ 
/* Disconnect the pipe instance. */
 
   if (! DisconnectNamedPipe(hPipe) ) 
      ErrorExit("Disconnect Named Pipe"); 
 
/* Call a subroutine to connect to the new client. */
 
   fPendingIO = ConnectToNewClient( 
      hPipe, 
      &oOverlap); 
 
   dwState = fPendingIO ? 
      CONNECTING_STATE : /* still connecting */
      READING_STATE;     /* ready to read */
} 


void PipeStuff() {
   DWORD cbBytes, dwErr;
   BOOL fSuccess; 

   Sleep(1000);
   fSuccess = WaitForSingleObject(oOverlap.hEvent, INFINITE);
   
   if (fPendingIO) 
      { 
         fSuccess = GetOverlappedResult( 
            hPipe, /* handle to pipe */
            &oOverlap, /* OVERLAPPED structure */
            &cbBytes,  /* bytes transferred */
            FALSE);    /* do not wait */
 
         switch (dwState) 
         { 
         /* Pending connect operation */
            case CONNECTING_STATE: 
               if (! fSuccess) 
                  ErrorExit("Connect Named Pipe (not CONNECTING_STATE)"); 
               dwState = READING_STATE; 
               break; 
 
            default: 
               ErrorExit("Invalid Pipe State"); 
         }  
      } 

      /* The pipe state determines which operation to do next. */

	  switch (dwState) 
      { 
      /* READING_STATE: 
         The pipe instance is connected to the client 
         and is ready to read a request from the client. */
 
         case READING_STATE: 
            fSuccess = ReadFile( 
               hPipe, 
               chBuf, 
               BUFSIZE, 
               &cbBytes, 
               &oOverlap); 
 
         /* The read operation completed successfully. */
 
            if (fSuccess && cbBytes != 0) 
            { 
               fPendingIO = FALSE; 
               dwState = WRITING_STATE; 
		        printf("child:\n");
        		printf("hStdin:%X\n", *(DWORD *)(char *)&chBuf);
				printf("hStdout:%X\n", *(DWORD *)((char *)&chBuf+sizeof(HANDLE)));
				printf("hStderr:%X\n", *(DWORD *)((char *)&chBuf+2*sizeof(HANDLE)));

			   break;
            } 
 
         /* The read operation is still pending. */
 
            dwErr = GetLastError(); 
            if (! fSuccess && (dwErr == ERROR_IO_PENDING)) 
            { 
               fPendingIO = TRUE; 
			   break;
            } 
 
         /* An error occurred; disconnect from the client. */
 
            DisconnectAndReconnect(); 
            break; 
   
         default: 
            ErrorExit("Invalid Pipe State"); 
      } 
}

void testProcess(char *filename, DWORD flag) {
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );

   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO);    

   sprintf(chBuf,"%s %d",filename,hPipe);
   if (!CreateProcess(filename, 
      chBuf, /* command line */
      NULL,          /* process security attributes */
      NULL,          /* primary thread security attributes */
      TRUE,          /* handles are inherited */
      flag,          /* creation flags*/
      NULL,          /* use parent's environment */
      NULL,          /* use parent's current directory */
      &siStartInfo,  /* STARTUPINFO pointer */
      &piProcInfo))  /* receives PROCESS_INFORMATION */
   ErrorExit("Create process failed"); 
   PipeStuff();   
}

void testFailedPipedProcess(char *filename, DWORD flag) {

   DWORD dwRead, dwWritten, resR, resW; 
   BOOL fSuccess; 
   BOOL inherit;
 
   HANDLE hSaveStdin, hSaveStdout;

   PROCESS_INFORMATION piProcInfo; 
   STARTUPINFO siStartInfo; 
 
   hSaveStdout = GetStdHandle(STD_OUTPUT_HANDLE);
   hSaveStdin = GetStdHandle(STD_INPUT_HANDLE);
 
/* Now create the child process. */
 
/* Set up members of the PROCESS_INFORMATION structure. */
 
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );
 
/* Set up members of the STARTUPINFO structure. */
 
   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO); 
   siStartInfo.hStdInput = hSaveStdin;
   siStartInfo.hStdOutput = hSaveStdout; 
   siStartInfo.hStdError = hSaveStdout;
   siStartInfo.dwFlags = STARTF_USESTDHANDLES;
/* Create the child process. */
 
   sprintf(chBuf,"%s %d",filename,hPipe);
   if (!CreateProcess(filename, 
      chBuf,         /* command line */
      NULL,          /* process security attributes */
      NULL,          /* primary thread security attributes */
      TRUE,          /* handles are inherited */
flag, /* creation flags */
      NULL,          /* use parent's environment */
      NULL,          /* use parent's current directory */
      &siStartInfo,  /* STARTUPINFO pointer */
      &piProcInfo))  /* receives PROCESS_INFORMATION */
   ErrorExit("Create process failed"); 

   PipeStuff();   
}

void testPipedProcess(char *filename, DWORD flag) {

   DWORD dwRead, dwWritten, resR, resW; 
   CHAR chBuf[BUFSIZE]; 
   CHAR chBuf2[BUFSIZE]; 
   BOOL fSuccess; 
   BOOL inherit;
 
   HANDLE hSaveStdin, hSaveStdout,
        hReadFromChildTmp, hReadFromChild, hChildInput,
        hChildOutput,hWriteToChild,hWriteToChildTmp;

   PROCESS_INFORMATION piProcInfo; 
   STARTUPINFO siStartInfo; 

   SECURITY_ATTRIBUTES sa;

   // Set up the security attributes struct.
   sa.nLength= sizeof(SECURITY_ATTRIBUTES);
   sa.lpSecurityDescriptor = NULL;
   sa.bInheritHandle = TRUE;

   inherit = FALSE;
   
   /* Set the bInheritHandle flag so pipe handles are inherited. 
 
    The steps for redirecting child process's STDOUT: 
        1. Save current STDOUT, to be restored later. 
        2. Create anonymous pipe to be STDOUT for child process. 
        3. Set STDOUT of the parent process to be write handle to 
           the pipe, so it is inherited by the child process. 
        4. Create a noninheritable duplicate of the read handle and
           close the inheritable read handle. 
*/ 

/* Save the handle to the current STDOUT. */
 
   hSaveStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
 
/* Create a pipe for the child process's STDOUT. */

   if (! CreatePipe(&hReadFromChildTmp, &hChildOutput, &sa, 0)) 
      ErrorExit("Stdout pipe creation failed\n"); 
 
/* Create noninheritable read handle and close the inheritable read 
   handle.  */

    fSuccess = DuplicateHandle(GetCurrentProcess(), hReadFromChildTmp,
        GetCurrentProcess(), &hReadFromChild, 0,
        inherit,
        DUPLICATE_SAME_ACCESS);
    if( !fSuccess )
        ErrorExit("DuplicateHandle failed");
    CloseHandle(hReadFromChildTmp);

/* Save the handle to the current STDIN. */
 
   hSaveStdin = GetStdHandle(STD_INPUT_HANDLE); 
 
/* Create a pipe for the child process's STDIN. */
 
   if (! CreatePipe(&hChildInput, &hWriteToChildTmp, &sa, 0)) 
      ErrorExit("Stdin pipe creation failed\n"); 
 
/* Duplicate the write handle to the pipe so it is not inherited. */
 
   fSuccess = DuplicateHandle(GetCurrentProcess(), hWriteToChildTmp, 
      GetCurrentProcess(), &hWriteToChild, 0, 
      inherit,                  
      DUPLICATE_SAME_ACCESS); 
   if (! fSuccess) 
      ErrorExit("DuplicateHandle failed"); 
 
   CloseHandle(hWriteToChildTmp); 
 
/* Now create the child process. */
 
/* Set up members of the PROCESS_INFORMATION structure. */
 
   ZeroMemory( &piProcInfo, sizeof(PROCESS_INFORMATION) );
 
/* Set up members of the STARTUPINFO structure. */
 
   ZeroMemory( &siStartInfo, sizeof(STARTUPINFO) );
   siStartInfo.cb = sizeof(STARTUPINFO); 
   siStartInfo.hStdInput = hChildInput;
   siStartInfo.hStdOutput = hChildOutput; 
   siStartInfo.hStdError = hChildOutput;
   siStartInfo.dwFlags = STARTF_USESTDHANDLES;
/* Create the child process. */
 
   sprintf(chBuf,"%s %d pipe",filename,hPipe);
   if (!CreateProcess(filename, 
      chBuf,         /* command line */
      NULL,          /* process security attributes */
      NULL,          /* primary thread security attributes */
      TRUE,          /* handles are inherited */
flag, /* creation flags */
      NULL,          /* use parent's environment */
      NULL,          /* use parent's current directory */
      &siStartInfo,  /* STARTUPINFO pointer */
      &piProcInfo))  /* receives PROCESS_INFORMATION */
   ErrorExit("Create process failed"); 

   CloseHandle(hChildInput); 
   CloseHandle(hChildOutput); 

   strcpy(chBuf, "trassa");
   dwRead = strlen(chBuf) + 1;
   chBuf2[0] = 0;

   resW = WriteFile(hWriteToChild, chBuf, dwRead, &dwWritten, NULL);
   resR = ReadFile( hReadFromChild, chBuf2, BUFSIZE, &dwRead, NULL);
 
   CloseHandle(piProcInfo.hProcess);
   TerminateThread(piProcInfo.hThread, 0); 
   if (strcmp(chBuf, chBuf2)) {
      printf("WRITE RESULT: %d\n",resW);
      printf("WRITTEN: %d bytes\n",dwWritten);
      printf("READ RESULT: %d\n",resR);
      printf("READ: %d bytes\n",dwRead);
      ErrorExit("Input output mismatched\n"); 
   }
   else {
      printf("INPUT: %s\n",chBuf);
      printf("OUTPUT: %s\n",chBuf2);
   }

   PipeStuff();   
}

DWORD main(int argc, char *argv[]) 
{ 
   HANDLE hStdin, hStdout, hStderr;
   BOOL fSuccess; 
   DWORD dwWrite;

   CHAR chBuf[BUFSIZE]; 
   DWORD dwRead, dwWritten; 

   hStdin = GetStdHandle(STD_INPUT_HANDLE); 
   hStdout = GetStdHandle(STD_OUTPUT_HANDLE); 
   hStderr = GetStdHandle(STD_ERROR_HANDLE); 

   if (argc > 1) {
        hPipe=CreateFile(PipeName,GENERIC_READ | GENERIC_WRITE,0,NULL
          ,OPEN_EXISTING,0,NULL);
        if(hPipe==INVALID_HANDLE_VALUE)
           ErrorExit("Opening pipe failed"); 

	dwMode = PIPE_READMODE_BYTE; 

	fSuccess = SetNamedPipeHandleState( 
      hPipe,    /* pipe handle */
      &dwMode,  /* new pipe mode */
      NULL,     /* don't set maximum bytes */
      NULL);    /* don't set maximum time */

   if (!fSuccess) 
      ErrorExit("SetNamedPipeHandleState"); 
        memcpy(chBuf, &hStdin, sizeof(HANDLE));
        memcpy(chBuf+sizeof(HANDLE), &hStdout, sizeof(HANDLE));
        memcpy(chBuf+2*sizeof(HANDLE), &hStderr, sizeof(HANDLE));
        WriteFile(hPipe,chBuf,3*sizeof(HANDLE),&dwWrite,NULL);
 
        if (argc > 2) {
            // Read from standard input. 
            fSuccess = ReadFile(hStdin, chBuf, BUFSIZE, &dwRead, NULL); 
            if (! fSuccess || dwRead == 0) 
                ExitProcess(0);
 
            // Write to standard output. 
            fSuccess = WriteFile(hStdout, chBuf, dwRead, &dwWritten, NULL); 
            if (! fSuccess) 
                ExitProcess(0);
        }
	else {
		printf("This is child's output\n");
	}
        ExitProcess(0);
   }

   printf("parent:\n");
   printf("hStdin:%X\n", hStdin);
   printf("hStdout:%X\n", hStdout);
   printf("hStderr:%X\n", hStderr);


   fSuccess = DuplicateHandle(GetCurrentProcess(), GetStdHandle(STD_INPUT_HANDLE),
        GetCurrentProcess(), &hStdin , 0,
        TRUE,
        DUPLICATE_SAME_ACCESS);
   if( !fSuccess )
        ErrorExit("DuplicateHandle failed");

    fSuccess = DuplicateHandle(GetCurrentProcess(), GetStdHandle(STD_OUTPUT_HANDLE),
        GetCurrentProcess(), &hStdout , 0,
        TRUE,
        DUPLICATE_SAME_ACCESS);
    if( !fSuccess )
        ErrorExit("DuplicateHandle failed");

    fSuccess = DuplicateHandle(GetCurrentProcess(), GetStdHandle(STD_ERROR_HANDLE),
        GetCurrentProcess(), &hStderr , 0,
        TRUE,
        DUPLICATE_SAME_ACCESS);
    if( !fSuccess )
        ErrorExit("DuplicateHandle failed");

   SetStdHandle(STD_INPUT_HANDLE, hStdin); 
   SetStdHandle(STD_OUTPUT_HANDLE, hStdout); 
   SetStdHandle(STD_ERROR_HANDLE, hStderr); 

   printf("\nparent duplicates:\n");
   printf("hStdin:%X\n", hStdin);
   printf("hStdout:%X\n", hStdout);
   printf("hStderr:%X\n", hStderr);

ZeroMemory( &oOverlap, sizeof(OVERLAPPED) );
   
   oOverlap.hEvent = CreateEvent( 
         NULL,    /* no security attribute */
         TRUE,    /* manual-reset event */
         TRUE,    /* initial state = signaled */
         NULL);   /* unnamed event object */

   if( oOverlap.hEvent == NULL )
        ErrorExit("Create event failed");
   
   hPipe= CreateNamedPipe( 
         PipeName,            /* pipe name  */
         PIPE_ACCESS_DUPLEX |     /* read/write access */
         FILE_FLAG_OVERLAPPED,    /* overlapped mode */
         PIPE_TYPE_BYTE |      /* message-type pipe */
         PIPE_READMODE_BYTE |  /* message-read mode */
         PIPE_WAIT,               /* blocking mode */
         INSTANCES,               /* number of instances */
         BUFSIZE,                 /* output buffer size */
         BUFSIZE,                 /* input buffer size */
         PIPE_TIMEOUT,            /* client time-out */
         NULL);                   /* no security attributes */
   
   if( hPipe == INVALID_HANDLE_VALUE )
        ErrorExit("Create pipe failed");
      fPendingIO = ConnectToNewClient( 
         hPipe, 
         &oOverlap); 
 
      dwState = fPendingIO ? 
         CONNECTING_STATE : /* still connecting */
         READING_STATE;     /* ready to read */

   printf("\nCreating new process\n");
   testProcess(argv[0],0);

   DisconnectAndReconnect(); 

   printf("\nCreating new CONSOLE process\n");
   testProcess(argv[0], CREATE_NEW_CONSOLE);

   DisconnectAndReconnect(); 

   printf("\nCreating new DETACHED process\n");
   testProcess(argv[0],DETACHED_PROCESS);

   DisconnectAndReconnect(); 

   printf("\nCreating PIPED new process\n");
   testPipedProcess(argv[0], 0);

   DisconnectAndReconnect(); 

   printf("\nCreating PIPED new CONSOLE process\n");
   testPipedProcess(argv[0], CREATE_NEW_CONSOLE);

   DisconnectAndReconnect(); 

   printf("\nCreating PIPED new DETACHED process\n");
   testPipedProcess(argv[0],DETACHED_PROCESS);

   DisconnectAndReconnect(); 

   printf("\nCreating FAILED PIPED new process\n");
   testFailedPipedProcess(argv[0], 0);

   DisconnectAndReconnect(); 

   printf("\nCreating FAILED PIPED new CONSOLE process\n");
   testFailedPipedProcess(argv[0], CREATE_NEW_CONSOLE);

   DisconnectAndReconnect(); 

   printf("\nCreating FAILED PIPED new DETACHED process\n");
   testFailedPipedProcess(argv[0],DETACHED_PROCESS);

	  if (! DisconnectNamedPipe(hPipe) ) 
		ErrorExit("DisconnectNamedPipe");
	  CloseHandle(hPipe);

	  CloseHandle(oOverlap.hEvent);
	  
   ExitProcess(0);
}  

